% Simplified data container
% behave to some degree like a regular matrix
% Input (squared) proximity matrix
% N - number of points accessible in Data
classdef DataContainer < matlab.mixin.Copyable
    properties (GetAccess='public',SetAccess = 'private')
        bInitialized    = false;
    end

    properties (GetAccess='private', SetAccess = 'protected')
        N               = 0;
        M               = 0;
        Data            = [];        
    end
    
    methods
        % constructor
        function obj = DataContainer(Data,varargin)
            if(iscell(Data))
                Data = Data{1};
            end
            if(nargin==1)
                obj.Data            = Data;
                [obj.N,obj.M]       = size(Data);
            else
                obj.N = varargin{1};
                obj.M = varargin{2};
                obj.Data = Data;
            end
            obj.bInitialized    = true;
        end

        % returns the parameter (not data matrix) of the binded container
        function oCell = GetParameter(me)
            oCell = {me.Data};
        end
                       
        function ret = mtimes(A,B)
            X = [];
            bLeft = false;
            if(isnumeric(A))
                bLeft = true;
                me = B;
                X  = A;
            else
                me = A;
                X  = B;
            end
            
            if(~bLeft)
                if(isvector(X))
                    ret = me.GetData()*X(:);
                else
                    M   = size(X,2); % 
                    Temp = me.GetData();
                    ret  = zeros(size(Temp,1),M);
                    for(k=1:M)
                        ret(:,k) = Temp * X(:,k);
                    end
                end
            else
                if(isvector(X))
                    ret = X(:)'*me.GetData();
                else
                    M    = size(X,1); %                     
                    Temp = me.GetData();
                    N    = size(Temp,2);
                    ret  = zeros(M,N);
                    for(k=1:M)
                        ret(k,:) = X(k,:) * Temp;
                    end
                end
                
            end
        end
        
        function ret = sum(me,varargin)
            if(nargin>2)
                dim = varargin{1}
                switch dim
                    case 1
                        ret = sum(me.GetData,1);
                    otherwise
                        ret = sum(me.GetData,2);
                end
            else
                ret = sum(me.GetData,1);
            end
        end
        
        function ret   = mean(me)
            ret = mean(me.GetData());
        end
        
        function ret   = std(me)
            ret = std(me.GetData());
        end
        
        function ret   = var(me)
            ret = var(me.GetData());
        end

        function ret   = full(me)
            ret = full(me.GetData());
        end
        
        function ret = GetData(me,X,Y)            
            if(~exist('X','var'))
                ret = me.Data;
            else
                ret = me.Data(X,Y);
            end
        end
        
        function [N,M] = size(me,varargin)
            if(nargin>1)
                switch varargin{1}
                    case 1,
                        N = me.N;
                    case 2
                        N = me.M;
                end
            else
                N = me.N;
                M = me.M;
            end
        end
            
        function Value = subsref(me,Indices)
            Value = [];
            NumInd = numel(Indices);
            if(NumInd==1) % direct access 
                Subs       = Indices.subs;
                NumIndSubs = numel(Subs);
                switch Indices.type
                    case '()'     % indexing                        
                        if(NumIndSubs==2)
                            Value = me.CalculateValue(Subs{1},Subs{2});
                        else
                            ME = MException('Class:DataContainer:Error','number of parameters must be two');
                            throw(ME);
                        end
                    case '.'     % class member
                        Value = me.(Subs);
                end
            else % should be a function call
                if(strncmp(Indices(1).type,'.',1))   
                    Value = builtin('subsref', me, Indices); 
                end
            end
        end       
        
        function Copy = minus(me,vec)
            Copy = copyElement(me);
            Copy.Data = bsxfun(@minus,Copy.GetData(),vec);
        end

        function Copy = plus(me,vec)
            Copy = copyElement(me);
            Copy.Data = bsxfun(@plus,Copy.GetData(),vec);
        end             
        
        function Copy = rdivide(me, vec)
            Copy = copyElement(me);
            Copy.Data = bsxfun(@rdivide,Copy.GetData(),vec);
        end
        
        % input 
        % X - index vector 
        % Y - index vector 
        function Value = CalculateValue(me,X,Y)       
            X = X(:);
            Y = Y(:);   
            Value = me.GetData(X,Y);
        end
    end
    
     methods(Access = protected)
       % Override copyElement method:
      function cpObj = copyElement(obj)
         % Make a shallow copy of all four properties
         cpObj = copyElement@matlab.mixin.Copyable(obj);      
      end
   end
end