%%%%%%%%%%%%
%
% Algorithm to calculate the core set of a set of points
% with probabilistic sampling (see JMLR, Tsang & Kwok, 2005
% for a core vector machine
%
%
% 
% IN:
%
% - fKPtr   - pointer to a kernel function e.g. fKPtr = @(X,Y)K(X,Y) - for a kernel matrix K
% - epsilon - eps of the minimum enclosing ball problem (MEB)
% - vDiags  - pre-calculated K(x,x) values for each x
%
% OUT:
% - S - set of core set points 
% - R - radius of the core set
% - vAlpha - loadings of the QP solution for S
%
% Note:
% 
% implemented by Frank-M. Schleif <fschleif@techfak.uni-bielefeld.de>
% 2011-2014
%
%% 
function [S,R,vAlpha] = core_set_one_two(fKPtr,epsilon,vDiags,vDeltas,kappa) % using only the kernel

% check input arguments
if(~exist('epsilon','var') || epsilon <= 0)% || epsilon > 1)
    epsilon = 0.1;   
end

% provide alternative quadratic-progr. solver if available
% if(1 && exist('qpas','file')==3) % spm-qpc newcastle library available
%     fPtrMinimalEnclosingBall = @(X,x0,Deltas)MinimalEnclosingBallSPM(X,[],Deltas,kappa); % x0 is a dummy - warm start is not supported by spm-qpc
% else
%     fPtrMinimalEnclosingBall = @(X,x0,Deltas)MinimalEnclosingBall(X,x0,Deltas,kappa);
% end
options = struct('Diagnostics','off','Algorithm','active-set','Display','off');
vDiags = vDiags(:);
N      = numel(vDiags);

%% init 
fDissFull    = @(X,Y)max(0,bsxfun(@minus,bsxfun(@plus,vDiags(X),vDiags(Y)'),2*max(0,fKPtr(X,Y))));
fMEBDistance = @(vDiags,vAlpha,mK,mKSS,vDelta)(vDiags' -2 * vAlpha' * mK + vAlpha' * mKSS * vAlpha + vDelta' );  
dTolerance   = (1+epsilon).^2; % for the epsilon problem

SFull        = [1:N]; 
if(0)
    S            = randsample(N,2)';                                              % choose 2 random points (could be done more clever - but the cluster maybe a mix - so be careful)
    SFull        = SFull(SFull~=S(1) & SFull~=S(2));                              % remove new added entry from SFull
    vAlpha       = 0.5*ones(2,1);                                                 % default loadings in case of 2 points (initial center)
    mKSS         = fKPtr(S,S);                                                    % kernel matrix
    R            = (vAlpha' * (diag(mKSS)+vDeltas(S)) -vAlpha' * mKSS * vAlpha);   % thats R^2 - initial radius
else
    S            = 1;%randsample(N,1);                                              % choose 1 random point
    SFull        = SFull(SFull~=S(1));                                           % remove new added entry from SFull
    S(2)         = argmax(fDissFull(S(1),SFull)');                               % take point furthest away
    SFull        = SFull(SFull~=S(2));
    vAlpha0      = 0.5*ones(2,1); % dummy - not used
    mKSS         = fKPtr(S,S);

    mKSS   = double(0.5*(mKSS+mKSS'));
    Ns     = numel(S);
    f      = zeros(Ns,1); 
    ones1N = ones(1,Ns);
    zerosN1= zeros(Ns,1);
    onesN1 = ones(Ns,1);
    %vAlpha = qpas(2*mKSS,f,[],[],ones1N,1,zerosN1,onesN1); % spm-qpc (newcastle) library
    %vAlpha = quadprog(2*mKSS,f,[],[], ones1N,1, zerosN1,onesN1,vAlpha0,options); % here we use the dual formulation including a kernel (here linear, euclidean)
    
    vAlpha = qpip(2*mKSS,f,[],[],ones1N,1,zerosN1,onesN1); % spm-qpc (newcastle) library    
    vDeltasS = vDeltas(S);
    R      = (vAlpha' * (diag(mKSS)+vDeltasS(:)) -vAlpha' * mKSS * vAlpha);    
    %[R,vAlpha]   = fPtrMinimalEnclosingBall(mKSS,vAlpha0,vDeltas(S));                                     % update radius (with warm start alpha if available     
end
if(numel(find(vAlpha>1e-10))<2)
    warning('reset initial point')
    SFull        = [1:N]; 
    S            = randsample(N,2)';                                              % choose 2 random points (could be done more clever - but the cluster maybe a mix - so be careful)
    SFull        = SFull(SFull~=S(1) & SFull~=S(2));                              % remove new added entry from SFull
    vAlpha       = 0.5*ones(2,1);                                                 % default loadings in case of 2 points (initial center)
    mKSS         = fKPtr(S,S);                                                    % kernel matrix
    vDeltasS = vDeltas(S);
    R            = (vAlpha' * (diag(mKSS)+vDeltasS(:)) -vAlpha' * mKSS * vAlpha);   % thats R^2 - initial radius
end
SInit        = S;
if(isempty(SFull))
    return;
end

%% core set calculation - inital step
iProbSampleSize = 59;
%iProbSampleSize = length(SFull);
vRandSampling= randsample(length(SFull),min(length(SFull),iProbSampleSize));

mK           = fKPtr(S,SFull(vRandSampling));                                                % kernel of core points to all out of radius points
mKSS         = fKPtr(S,S);      % SFull is without S    
vDistances   = fMEBDistance(vDiags(SFull(vRandSampling)),vAlpha, mK, mKSS,vDeltas(SFull(vRandSampling)));                  % get distances for MEB
%% loop  until MEB is found
iCounter     = 2;
Rold         = R;
%bQPIP        = true;
while(~isempty(SFull))      
     vOutOfRadius     = find(vDistances > dTolerance * R);                    % identify points out of MEB
%     [numel(vOutOfRadius), min(vDistances)]
     if(isempty(vOutOfRadius))                                                % if there are not to many it is still ok
         break;
     end     
     [~,I]            = max(vDistances(vOutOfRadius));                        % largest point away - original core-set heuristic
     iNewCoreIndex    = SFull(vRandSampling(vOutOfRadius(I)));                % identify new coreset point
     iCounter         = iCounter+1;
     SLast            = S;
     S                = [S, iNewCoreIndex];                                   % add to core-set
     SFull            = SFull(SFull~=iNewCoreIndex);                          % remove new added entry from SFull and all which are in the circle
     mK               = fKPtr(S,S(end));
     mKSS             = [mKSS mK(1:end-1);mK'];
     %mKSS             = fKPtr(S,S);
     vAlphaLast       = vAlpha;
     vAlpha0          = [vAlpha; 0.5]';
     vAlpha0          = vAlpha0./sum(vAlpha0);         

     Ns     = numel(S);
     f      = zeros(Ns,1); 
     ones1N = ones(1,Ns);
     zerosN1= zeros(Ns,1);
     onesN1 = ones(Ns,1);
     mKSS   = double(0.5*(mKSS+mKSS'));
     %vAlpha = qpip(2*mKSS,f,[],[],ones1N,1,zerosN1,onesN1); % spm-qpc (newcastle) library
     
%      if(~bQPIP)
%         vAlpha = qpas(2*mKSS,f,[],[],ones1N,1,zerosN1,onesN1); % spm-qpc (newcastle) library
%      end
     %vAlpha = quadprog(2*mKSS,f,[],[], ones1N,1, zerosN1,onesN1,vAlpha0,options); % here we use the dual formulation including a kernel (here linear, euclidean)         
     %vAlpha = qpip(2*mKSS,f,[],[],ones1N,1,zerosN1,onesN1); % spm-qpc (newcastle) library
 %    if(bQPIP || sum(abs(vAlpha0(:)-vAlpha(:)))>1)        
         vAlpha = qpip(2*mKSS,f,[],[],ones1N,1,zerosN1,onesN1); % spm-qpc (newcastle) library
  %       bQPIP = true;
     %    % get a final valid alpha and stop
     %    break;
%     end
     
     %[sum(abs(vAlpha0-vAlpha1')) sum(abs(vAlpha0-vAlpha2'))]
     %vAlpha = vAlpha1;
     %sum(abs(vAlpha1-vAlpha2))
     %vAlpha = lsqlin(mKSS,1/Ns*ones(Ns,1),ones1N,1,[],[],zerosN1,onesN1,vAlpha0,options);
     %vAlpha = quadprog(2*mKSS,f,[],[], ones1N,1, zerosN1,onesN1,vAlpha0,options); % here we use the dual formulation including a kernel (here linear, euclidean)     
     vDeltasS = vDeltas(S);     
     R      = (vAlpha' * (diag(mKSS)+vDeltasS(:)) -vAlpha' * mKSS * vAlpha);

     %[R,vAlpha]       = fPtrMinimalEnclosingBall(mKSS,vAlpha0,vDeltas(S));                                     % update radius (with warm start alpha if available     
     if(any(vAlpha>1+2*eps) || any(vAlpha+2*eps<0) || (sum(vAlpha)-1)>1e-8)
          if(numel(find(vAlpha>1e-10))<2)
              warning('try active set approach')
              [R,vAlpha] = MinimalEnclosingBall(mKSS,vAlpha0,vDeltas(S),kappa);              
          end 
          if(numel(find(vAlpha>1e-10))<2)
            warning('optimization failed - use last valid solution');
            vAlpha = vAlphaLast;
            S      = SLast;
          end
          break;
     end
     vRelevant           = find(vAlpha>1e-10);
     vAlpha              = vAlpha(vRelevant);
     vAlpha              = vAlpha./sum(vAlpha);
     S                   = S(vRelevant);
     mKSS                = mKSS(vRelevant,vRelevant);
     vRandSampling       = randsample(N-iCounter,min(N-iCounter,iProbSampleSize)); 
     if(~isempty(vRandSampling) && ~isempty(SFull))
        mK               = fKPtr(S,SFull(vRandSampling));                                              % kernel of core points to all out of radius points       
        vDistances   = max(fMEBDistance(vDiags(SFull(vRandSampling)),vAlpha, mK, mKSS,vDeltas(SFull(vRandSampling))),0); % calculate new distances
        iProbSampleSize = numel(vRandSampling);
     end
     if(abs(Rold - R)<0.5*1e-6)
         numel(S)
         break;
     end     
     Rold = R;
end
% length(vAlpha)
% remove entries which vanished from the optimization - may happen 
if(sum(isnan(vAlpha))==0)
    S      = S(vAlpha>eps);
    vAlpha = vAlpha(vAlpha>eps);
    R      = sqrt(R); % return only R not R^2
else
    warning('kernel is invalid - solution may not be optimal - use last valid solution')
    vAlpha = vAlpha0;
    S      = S(vAlpha>eps);
    vAlpha = vAlpha(vAlpha>eps);
    R      = sqrt(R); % return only R not R^2    
end
if(numel(S)<2) % if the optimizer fails provide a backup
    warning('use backup')
    S       = SInit;
    vAlpha  = 0.5*ones(2,1);                                                 % default loadings in case of 2 points (initial center)
    mKSS    = fKPtr(S,S);                                                    % kernel matrix
    R       = sqrt(vAlpha' * (diag(mKSS)+vDeltas(S)) -vAlpha' * mKSS * vAlpha);           % thats R - initial radius
end

function [R,vAlpha] = MinimalEnclosingBall(mK,vAlpha0,Deltas,kappa)
    N = size(mK,1);
    f = zeros(N,1); 
    %f = -(diag(mK)+Deltas(:)-kappa);
    options = struct('Diagnostics','off','Algorithm','active-set','Display','off');
    mK = double(0.5*(mK+mK'));
    vAlpha = quadprog(2*mK,f,[],[], ones(1,N),1, zeros(N,1),ones(N,1),vAlpha0,options); % here we use the dual formulation including a kernel (here linear, euclidean)
    R = (vAlpha' * (diag(mK)+Deltas(:)) -vAlpha' * mK * vAlpha);

function [R,vAlpha] = MinimalEnclosingBallSPM(mK,x0,Deltas,kappa)
    N = size(mK,1);
    f = zeros(N,1); 
    %f = -(diag(mK)+Deltas(:)-kappa); % thats actually irrelevant because its just a constant in the costs
    Z = zeros(N,1);
    mK = double(0.5*(mK+mK'));
    vAlpha = qpip(2*mK,f,[],[],ones(1,N),1,Z,ones(N,1)); % spm-qpc (newcastle) library
    R = (vAlpha' * (diag(mK)+Deltas(:)) -vAlpha' * mK * vAlpha);
   
    
    